
package algo;

import java.io.IOException;
import java.io.Serializable;
import java.io.FileNotFoundException;
import java.util.ArrayList;

import comm.ChannelFactory;
import comm.ExecutableChannels;
import comm.SocketChannel;
import comm.RemoteExecutable;
import thread.Deque;
import util.Logger;

/**
 * A parallel algorithm for an euclidean 2d TSP Problem
 */

public class DistTSP implements TSPInf {

    protected Graph graph;
    protected Point[] points;
     protected DistBestStore best;

    protected int number = 0;
    protected String hostfile = "machines";
    protected int port = -1;
    protected String master = "localhost";
    protected ChannelFactory cf;
    protected DistRunTSP[] threads;

    private static Logger logger = new Logger(2);

    public DistTSP(Point[] p, int th) {
        this.points = p;
        graph = new PlaneGraph(p);
        number = th;
        threads = null;
        // distribute code
        ExecutableChannels ec = null;
        try {
            ec = new ExecutableChannels( hostfile );
            ec.open( number );
            master = ec.getMasterHost();
            port = ec.getMasterPort();
            for ( int i = 0; i < number; i++ ) {
                ec.send(i, new DistParClientTSP(master,port,points) );
            }
            logger.info("code distributed");
        } catch (FileNotFoundException e) {
            e.printStackTrace();
            return;
        } catch (IOException e) {
            e.printStackTrace();
            return;
        }
        if ( ec != null ) {
           ec.close();
        }
        cf = new ChannelFactory(port);
    }


   public long getIterations() {
        if ( threads == null ) {
            return 0l;
        }
        long iter = 0l;
        for ( int i = 0; i < threads.length; i++ ) {
            if ( threads[i] != null ) {
               iter += threads[i].getIterations();
            }
        }
        return iter;
    }

    public long getMaxIterations() {
        if ( threads == null ) {
            return Long.MAX_VALUE;
        }
        long maxiter = Long.MAX_VALUE;
        for ( int i = 0; i < threads.length; i++ ) {
            long mi = threads[i].getMaxIterations();
            if ( maxiter > mi ) {
                maxiter = mi;
            }
        }
        return maxiter;
    }

    public long setMaxIterations(long m){
        if ( threads == null ) {
            return m;
        }
        long maxiter = Long.MAX_VALUE;
        for ( int i = 0; i < threads.length; i++ ) {
            long mi = threads[i].setMaxIterations(m);
            if ( maxiter > mi ) {
                maxiter = mi;
            }
        }
        return maxiter;
    }


    public Path actualBest() {
        if ( best == null ) {
            return null;
        }
        return best.getPath();
    }

    public void setBest(Path b) {
        if ( b == null ) {
           return;
        }
        if ( best == null ) {
           // best.setPath( b );
           return;
        }
        if ( b.cost() < best.getPath().cost() ) {
           best.setPath( b );
        }
    }


    public Path getBest() {
        return getBest(Long.MAX_VALUE);
    }

    public Path getBest(long max) {
        Path bestPath = PathByteArray.standardPath(graph); // standard path
        if ( bestPath.length() <= 2 ) {
            return bestPath;
        }
        best = new DistBestStore( bestPath, 0, number );
        Path path = PathByteArray.onePath(graph); // one/empty path
        Deque globalStack = new Deque(10000);
        try {
            globalStack.push( path );
        } catch (InterruptedException e) {
            e.printStackTrace();
        }
        threads = new DistRunTSP[number];
        for ( int i = 0; i < number; i++ ) {
            SocketChannel sc = null;
            try {
                sc = cf.getChannel();
                //System.out.println("getChannel() = "+sc);
            } catch (InterruptedException e) {
                e.printStackTrace();
            }
            threads[i] = new DistRunTSP(globalStack,best,max,sc);
            threads[i].start();
        }
        try {
            for ( int i = 0; i < number; i++ ) {
                threads[i].join();
            }
        } catch (InterruptedException e) {
            e.printStackTrace();
        }
        cf.terminate();
        return best.getPath();
    }

    public void getBest(Path path) {
        System.out.println("method getBest() in DistTSP not implemented");
    }

}

class DistBestStore {

    private Path path;
    private int idlers;
    public final int threads;

    public DistBestStore(Path b, int i, int t) {
        path = b;
        idlers = i;
        threads = t;
    }

    public Path getPath() {
        return path;
    }

    public void setPath(Path p) {
        path = p;
    }

    public synchronized void incIdle() {
        idlers++;
    }

    public synchronized void decIdle() {
        idlers--;
    }

    public synchronized /*elsewhere*/ boolean allIdle() {
        return (idlers == threads);
    }
}


class DistRunTSP extends Thread {

    private DistBestStore best;
    private Deque globalStack;
    private long maxIter;
    private long iter;
    private long localWork;
    private long pushes;
    private long requests;
    private ArrayList stack;
    private SocketChannel comm;
    private static Logger logger = new Logger(2);

    public DistRunTSP(Deque s, 
                      DistBestStore b, 
                      long max, 
                      SocketChannel comm) {
        globalStack = s;
        best = b;
        maxIter = max;
        stack = new ArrayList(500);
        this.comm = comm;
    }


    public void run() {
        requests = 0;
        pushes = 0;
        Object o = null;
        Object m = null;
        Path p = null;
        Path w = null;
        TransportMessage dtc;
        while (true) {
            // protocol: receive/push/send      
            o = null;
            try {
                o = comm.receive();
                requests++;
                logger.debug("DistRunTSP receive "+comm+" request "+o);
            } catch (IOException e) {
               e.printStackTrace();
               //break;
            } catch (ClassNotFoundException e) {
               e.printStackTrace();
               break;
            }
            if ( ! (o instanceof TransportMessage) ) {
                break;
            }
            dtc = (TransportMessage)o;
            if ( dtc instanceof TransportMessageGetWork ) {
              //logger.debug("DistRunTSP search work");
              Object work = null;
              try {
                best.incIdle();
                if ( globalStack.empty() && best.allIdle() ) {
                   logger.debug("pushing nulls");
                   for ( int i = 0; i < best.threads; i++) {
                       globalStack.push(null);
                   }
                   //best.decIdle(); break; return;
                }
                //logger.debug("work blocking " + this + " " + comm);
                work = globalStack.get(); // blocking pop/get
                best.decIdle();
              } catch (InterruptedException e) {
                e.printStackTrace();
                break;
              }
              logger.debug("work of " + globalStack.size() 
                           + " found @ " + this + " = "+work);
              try {
                  if ( work == null ) {
                     //logger.debug("no more Work available");
                     comm.send( (TransportMessage)null );
                     break;
                  }
                  p = (Path)work;
                  dtc = new TransportMessage( best.getPath(),
                                              p,
                                              maxIter,
                                              globalStack.size() );
                  comm.send( dtc );
                  logger.debug("response sendWork = " + dtc);
              } catch (IOException e) {
                 e.printStackTrace();
                 break;
              }
            } else { // normal message
               p = dtc.best;
               w = dtc.work; 
               iter = dtc.iter;
               localWork = dtc.items;
               if ( w != null ) {
                  logger.debug(" put Part to global Stack "+w);
                  try {
                      globalStack.put( w );
                      pushes++;
                  } catch (InterruptedException e) {
                      e.printStackTrace();
                      break;
                  }
               }
               if ( p != null ) {
                  synchronized (best) {
                     if ( p.cost() < best.getPath().cost() ) {
                        best.setPath( p );
                     }
                  }
               }
            }
        } 
        logger.info("requests = " + requests
                  + " iterations = " + iter
                  + " pushes = " + pushes);
        comm.close();
    }

    public long getIterations() {
        if ( comm == null ) {
           return iter;
        }
        try {
            comm.send( new TransportMessage( best.getPath(),
                                             (Path)null,
                                             maxIter,
                                             globalStack.size()) );
        } catch (IOException e) {
            if ( logger.isDebugEnabled() ) {
               e.printStackTrace();
            }
            // maxIter = iter;
        }
        return iter;
    }

    public long getMaxIterations() {
        if ( comm == null ) {
           return maxIter;
        }
        try {
            comm.send( new TransportMessage( best.getPath(),
                                             (Path)null,
                                             maxIter,
                                             globalStack.size()) );
        } catch (IOException e) {
            e.printStackTrace();
            // maxIter = iter;
        }
        return maxIter;
    }

    public long setMaxIterations(long m){
        long x = maxIter;
        maxIter = m;
        if ( comm == null ) {
           return x;
        }
        try {
            comm.send( new TransportMessage( best.getPath(),
                                             (Path)null,
                                             maxIter,
                                             globalStack.size()) );
        } catch (IOException e) {
            e.printStackTrace();
            // maxIter = iter;
        }
        return x;
    }

}


/**
 * Objects of this class are to be send to a ExecutableServer
 */

class DistParClientTSP implements RemoteExecutable {

    protected String host;
    protected int port;
    protected Graph graph;
    protected Point[] points;
    protected DistParBestStore best;
    protected int number = 0;
    protected ArrayList stack;
    protected int count;
    protected int depth;
    protected long iter;
    protected long maxIter;
    protected long globalWork;
    protected int pushes;
    protected int runs;
    protected DistParRemoteCommTSP communi;
    private static Logger logger = null;

    public DistParClientTSP(String host, int port, Point[] points) {
	this.host = host;
	this.port = port;
	this.points  = points;
        iter = 0;
        maxIter = Long.MAX_VALUE;
        pushes = 0;
        runs = 0;
        globalWork = 0;
    }

    public Path actualBest() {
        return best.getPath();
    }

    public void setBest(Path b) {
        if ( b == null ) {
           return;
        }
        if ( best == null ) {
           // best.setPath( b );
           return;
        }
        synchronized ( best ) {
           if ( b.cost() < best.getPath().cost() ) {
              best.setPath( b );
           }
        }
    }


    public void run() {
        logger = new Logger(2);
        graph = new PlaneGraph(points);
        Path bestPath = PathByteArray.standardPath(graph); // standard path
        best = new DistParBestStore(bestPath);
        int ms = bestPath.maxsize();
        count = ms + 4; //*2 / +5 ?
        depth = ms - 10; // = 3 ?
        if ( depth < 3 ) {
            depth = 3;
        }
        stack = new ArrayList(500);
        globalWork = 0;

        communi = new DistParRemoteCommTSP(host,port,this);
        communi.start();
        logger.info("DistParClientTSP:" 
                  + " host = " + host
                  + " port = " + port
                  + " best = " + actualBest()
                   );

        //Object o;
        //Path p;
        //Path w;
        try {
            getParBest(); // blocking
        } catch (InterruptedException e) {
            e.printStackTrace();
        }
        logger.info("dist client runs = " + runs
                  + " iterations = " + iter);
        communi.terminate();
    }


    public void getParBest() throws InterruptedException {
        // stack.add( path );
        Object o;
        Path p;
        while ( true ) {
            if ( stack.size() > 0 ) {
               synchronized (stack) {
                  o = stack.remove( stack.size()-1 );
               }
               if ( o == null ) {
                  break; //return;
               }
               p = (Path)o;
               getBest(p); // blocking
            } else {
               if ( communi != null ) {
                  communi.sendGetWork();
               } else {
                   break;
               }
               //logger.debug("getParBest wait()");
               synchronized (stack) {
                  if ( stack.size() == 0 ) { // recheck hard
                     stack.wait();
                  }
               }
            }
        }
        logger.debug("getParBest terminated");
    }

    public void addEndMark() {
        synchronized ( stack ) {
           stack.add( null );
           stack.notify(); // wake up main
        }
    }

    public void putWork(Path path) {
        // termination ?
        if ( path == null ) {
           return;
        }
        synchronized (stack) {
           //int s = stack.size();
           runs++;
           stack.add( path );
           //if ( s == 0 ) {
           stack.notify();
           //}
        }
    }

    public Path getWork() {
        Path w = null;
        if ( stack.size() > count 
             && ( /*actualBest().length() <= depth &&*/ globalWork <= 5 ) ) {
           Object o = null;
           synchronized (stack) {
              if ( stack.size() > 0 ) {
                 o = stack.remove( 0 ); //stack.size()-1, depth first
              }
           }
           if ( o != null ) {
               w = (Path)o;
               pushes++;
           }
        }
        return w;
    }

    public boolean isWorkAvailable(int len) {
        //logger.debug("DistParRemoteCommTSP, isWorkAvailable "+globalWork 
        //           + " stack " + stack.size() 
        //           + " length " + actualBest().length() );
        if ( stack.size() > count 
             && ( len <= depth && globalWork <= 5 ) ) {
            communi.sendPutWork(); 
            return true;
        }
        return false;
    }

    public void getBest(Path path) {
        Path[] ps = path.nextPaths(); 
        iter += ps.length;
        double bestCost = best.getPath().cost();
        if ( path.length()+1 == path.maxsize() ) {
           for ( int i = 0; i < ps.length; i++ ) { // branch
               if ( ps[i].cost() < bestCost ) {
                 synchronized ( best ) {
                     if ( ps[i].cost() < best.getPath().cost() ) {
                        bestCost = ps[i].cost();
                        best.setPath( (Path)ps[i].clone() ); // evtl. error
                        logger.info("dist par best = " + ps[i]);
                   }
                 }
               }
           }
           return;
        }
        if ( iter >= maxIter ) {
            return;
        }
        if ( path.length()+5 >= path.maxsize() ) {
           for ( int i = 0; i < ps.length; i++ ) { // branch
               if ( ps[i].cost() < bestCost ) { // cut
                   //System.out.print("#");
                   getBestRec( ps[i].copyMax(), path.length()+1 ); // changes best and bestPath
                  //stack.add( ps[i] );
               }
           }
           return;
        }
        for ( int i = 0; i < ps.length; i++ ) { // branch
            if ( ps[i].cost() < bestCost ) { // cut
               // getBest( ps[i] ); // changes best and bestPath
               //System.out.print("-");
                synchronized ( stack ) {
                   stack.add( ps[i] );
                }
            }
        }
        if ( ps.length > 0 ) {
           isWorkAvailable( ps[0].length() );
        }
        return;
    }

    public void getBestRec(Path path, int depth) {
        double bestCost = best.getPath().cost();
        for ( int k = 0; k < path.maxsize()-depth; k++ ) { // branch
            Path ps = path.nextPath(k); 
            iter++;
            if ( depth+1 == path.maxsize() ) {
               if ( ps.cost() < bestCost ) {
                  synchronized ( best ) {
                     if ( ps.cost() < best.getPath().cost() ) {
                        bestCost = ps.cost();
                        best.setPath( (Path)ps.clone() ); // had error
                        logger.info("dist par best = " + ps);
                     }
                  }
               }
            }
            if ( iter >= maxIter ) {
               return;
            }
            if ( ps.cost() < bestCost ) { // cut
                getBestRec( ps, depth+1 ); // changes best and bestPath
                //stack.add( ps[i] );
            }
        }
        return;
    }

    public long getIterations() {
        return iter;
    }

    public long getMaxIterations() {
        return maxIter;
    }

    public long setMaxIterations(long m){
        long x = maxIter;
        maxIter = m;
        return x;
    }
    /*
    public long getGlobalWork() {
        return globalWork;
    }
    */
    public void setGlobalWork(long w){
        globalWork = w;
        return;
    }

}


class DistParBestStore implements Serializable {

    private Path path;

    public DistParBestStore(Path b) {
        path = b;
    }

    public Path getPath() {
        return path;
    }

    public void setPath(Path p) {
        path = p;
    }
}


class TransportMessage implements Serializable {
    final Path best;
    final Path work;
    final long iter;  // or maxIter
    final long items; // work items global / local

    public TransportMessage(Path b, Path w, long i, long wi) {
        best = b;
        work = w;
        iter = i;
        items = wi;
    }
    public TransportMessage(Object b, Object w, long i, long wi) {
        this((Path)b,(Path)w,i,wi);
    }

    public String toString() {
        return "DTP(" + best + "," + work + "," + iter + "," + items + ")";
    }
}

class TransportMessageGetWork extends TransportMessage {

    public TransportMessageGetWork() {
        super((Path)null,(Path)null,0l,0l);
    }

}


/**
 * Objects of this class communicate with the master server
 */

class DistParRemoteCommTSP extends Thread implements Serializable {

    protected String host;
    protected int port;
    protected SocketChannel comm = null;
    protected DistParClientTSP rem;
    private final static Logger logger = new Logger(2);


    public DistParRemoteCommTSP(String host, int port, DistParClientTSP rem) {
	this.host = host;
	this.port = port;
	this.rem = rem;

        ChannelFactory cf = new ChannelFactory();
        try {
            comm = cf.getChannel( host, port );
        } catch (IOException e) {
            e.printStackTrace();
        }
        cf.terminate();
    }

    public void terminate() {
        try { // wg isBestPath before comm.close()
            Thread.sleep(200); 
        } catch (InterruptedException e) { 
        }
        if ( comm != null ) {
           comm.close();
        }
        try { 
            while ( this.isAlive() ) {
                  //System.out.print(".");
                  this.interrupt(); 
                  this.join(100);
            }
            //System.out.println("TSPModelCommRemote terminated");
        } catch (InterruptedException e) { 
        }
    }

    public void run() {
        if ( comm == null ) {
            return;
        }
        TransportMessage dtc;
        Object m;
        Path p;
        Path w;
        long iter;
        long maxIter;
        long items;
        int mess = 0;
        while ( true ) {
              m = null;
              try {
                  m = comm.receive();
                  mess++;
              } catch (IOException e) {
                  if ( logger.isDebugEnabled() ) {
                     e.printStackTrace();
                  }
              } catch (ClassNotFoundException e) {
                  e.printStackTrace();
              }
              logger.debug("DistParRemoteCommTSP, receive m = " + m);
              dtc = null;
              p = null;
              w = null;
              if ( m instanceof TransportMessage ) {
                 dtc = (TransportMessage)m;
                 rem.setBest( dtc.best );
                 rem.putWork( dtc.work );
                 rem.setMaxIterations( dtc.iter );
                 rem.setGlobalWork( dtc.items );
                 //logger.debug("recieved work = " + dtc.work
                 //             + ", path = " + dtc.best);
              }
              Object response = null;
              response = new TransportMessage( rem.actualBest(),
                                               (Path)null,
                                               rem.getIterations(),
                                               rem.stack.size() );
              try {
                  if ( response != null ) {
                      //logger.debug("response to send = " + response);
                     comm.send( response );
                  }
              } catch (IOException e) {
                  //e.printStackTrace();
                  break;
              }
        }
        finalBest();
        logger.info("dist client comm messages = " + mess);
        // notify worker
        rem.addEndMark();
    }

    public void finalBest() {
        if ( comm == null ) {
            return;
        }
        Object response = null;
        try {
            response = new TransportMessage( rem.actualBest(),
                                             (Path)null,
                                             rem.getIterations(),
                                             rem.stack.size() );
            comm.send( response );
        } catch (IOException e) {
            //e.printStackTrace();
        }
    }

    public void sendGetWork() {
        try {
            logger.debug("sendGetWork " +comm);
            comm.send( new TransportMessageGetWork() );
        } catch (IOException e) {
            e.printStackTrace();
        }
    }

    public void sendPutWork() {
        try {
            Object r = null;
            r = new TransportMessage( rem.actualBest(),
                                      rem.getWork(),
                                      rem.getIterations(),
                                      rem.stack.size() );
            comm.send( r );
            logger.debug("sendPutPart = "+r);
        } catch (IOException e) {
            e.printStackTrace();
        }
    }

}
