
package algo;

import java.io.Serializable;

/**
 * A Path in a Graph implementd as Arrays of bytes. 
 */

public final class PathByteArray implements Path, Serializable {

    public final Graph graph;
    public final byte[] used;   // points in the path
    public final byte[] unused; // points not in the path
    public final double cost;
    public final int maxsize;
    public final int length;

    public PathByteArray(Graph g, byte[] u, byte[] uu, double c) {
        this(g, u, u.length, uu, c);
    }

    public PathByteArray(Graph g, byte[] u, int pos, byte[] uu, double c) {
        graph = g;
        used = u;
        unused = uu;
        cost = c;
        length = pos;
        maxsize = graph.size();
    }


    public Object clone() {
        byte[] u = new byte[ length ]; 
        System.arraycopy( used, 0, u, 0, length );
        return new PathByteArray( graph, u, length, unused, cost );
    }


    public double cost() {
        return cost;
    }

    public int length() {
        return length;
    }

    public int maxsize() {
        return maxsize;
    }


    /**
     * empty path, all n points unused.
     */

    public static PathByteArray emptyPath(Graph g) { 
        byte[] u  = new byte[ 0 ];
        byte[] uu = new byte[ g.size() ];
        for (byte i = 0; i < g.size(); i++ ){
            uu[i] = i;
        }
        return new PathByteArray(g, u, uu, Double.POSITIVE_INFINITY);
    }

    public static PathByteArray onePath(Graph g) { 
        byte[] u  = new byte[ 1 ];
        u[0] = (byte)0;
        byte[] uu = new byte[ g.size()-1 ];
        for (byte i = 0; i < g.size()-1; i++ ){
            uu[i] = (byte)(i+1);
        }
        return new PathByteArray(g, u, uu, Double.POSITIVE_INFINITY);
    }


    /**
     * standard path from point 0 to point n-1 to 0.
     */

    public static PathByteArray standardPath(Graph g) {
        byte[] u  = new byte[ g.size() ];
        byte[] uu = new byte[0];
        double c = 0.0;
        u[0] = (byte)0;
        for (byte i = 1; i < g.size(); i++ ){
            u[i] = i;
            c += cost( g.distance(i-1,i) );
        }
        // u[ g.size() ] = (byte)0;
        c += cost( g.distance(g.size()-1,0) ); //for circle
        PathByteArray p = new PathByteArray(g, u, uu, c);
        System.out.println("initial  = " + p);
        return p;
    }


    /**
     * return a copy of this with maximal used array.
     */

    public Path copyMax() {
        byte[] u = new byte[ maxsize ]; 
        System.arraycopy( used, 0, u, 0, length );
        return new PathByteArray( graph, u, length, unused, cost );
    }


    /**
     * get next possible path reusing used array.
     */

    public Path nextPath(int pos) {
        //System.out.print("pos = " + pos + " this = " + this);
        byte[] uu = new byte[ unused.length-1 ];
        byte m = 0;
        byte j = -1;
        for ( byte k = 0; k < unused.length; k++ ) {
            if ( k == pos ) { 
                 j = unused[k];
            } else {
                 uu[ m ] = unused[k]; m++;
            }
        }
        // j != -1
        used[ length ] = j;
        // compute new cost
        double c = 0.0;
        if ( cost < Double.POSITIVE_INFINITY ) {
           c = cost;
        }
        if ( length > 0 ) {
           int n1 = used[ length-1 ];
           c += cost( graph.distance(n1,j) );
        }
        // add circle cost
        if ( length+1 == maxsize ) {
           //System.out.print("#");
           int n0 = used[ 0 ];
           c += cost( graph.distance(j,n0) );
        }
        // new path shares used array, so do not use in threads
        return new PathByteArray( graph, used, length+1, uu, c ); 
    }


    /**
     * get array of next possible paths.
     */

    public Path[] nextPaths() {
        PathByteArray[] pr = new PathByteArray[ unused.length ];
        for ( byte i = 0; i < unused.length; i++ ) {
            byte[] uu = new byte[ unused.length-1 ];
            byte m = 0;
            byte j = -1;
            for ( byte k = 0; k < unused.length; k++ ) {
                if ( k == i ) { 
                    j = unused[k];
                } else {
                    uu[ m ] = unused[k]; m++;
                }
            }
            // j != -1
            byte[] u = new byte[ used.length+1 ];
            System.arraycopy( used, 0, u, 0, used.length );
            u[ used.length ] = j;
            // compute new cost
            double c = 0.0;
            if ( cost < Double.POSITIVE_INFINITY ) {
               c = cost;
            }
            if ( length > 0 ) {
               int l = used[ length-1 ];
               int k = j;
               c += cost( graph.distance(l,k) );
            }
            // add circle cost
            if ( length+1 == maxsize ) {
                //System.out.print("#");
               int n = used[ 0 ];
               int k = j;
               c += cost( graph.distance(k,n) );
            }
            pr[i] = new PathByteArray( graph, u, uu, c ); 
        }
        return pr;
    }


    public String toString() {
        StringBuffer s = new StringBuffer();
        s.append("[ ");
        for ( int i = 0; i < length; i++ ) {
            s.append( used[i] );
            if ( i+1 < length ) {
               s.append(",");
            }
        }
        s.append(" | cost="+cost);
	if ( unused.length > 0 ) {
           s.append(" | ");
           for ( int i = 0; i < unused.length; i++ ) {
               s.append( unused[i] );
               if ( i+1 < unused.length ) {
                  s.append(",");
               }
           }
	}
        s.append(" ]");
        return s.toString();
    }


    protected static double cost(double c) {
        if ( c >= 0.0 ) {
            return c;
        }
        double r = Math.log( c );
        for (int i = 0; i < 100; i++ ) {
            r *= c;
        }
        r = r/2.0;
        for (int i = 0; i < 100; i++ ) {
            r /= c;
        }
        r = r/2.0;
        return Math.exp( r );
    }

}
